import copy
import createrepo_c
import json
from gettext import gettext as _

from django.db import IntegrityError
from jsonschema import Draft7Validator
from rest_framework import serializers
from rest_framework.exceptions import NotAcceptable

from pulpcore.plugin.models import (
    AsciiArmoredDetachedSigningService,
    Remote,
)
from pulpcore.plugin.serializers import (
    ArtifactSerializer,
    ContentChecksumSerializer,
    DetailRelatedField,
    ModelSerializer,
    MultipleArtifactContentSerializer,
    NoArtifactContentSerializer,
    PublicationDistributionSerializer,
    PublicationSerializer,
    RepositorySyncURLSerializer,
    RelatedField,
    RemoteSerializer,
    RepositorySerializer,
    SingleArtifactContentUploadSerializer,
    validate_unknown_fields,
)

from pulp_rpm.app.advisory import hash_update_record
from pulp_rpm.app.constants import CHECKSUM_CHOICES, CHECKSUM_TYPES

from pulp_rpm.app.fields import (
    UpdateCollectionPackagesField,
    UpdateReferenceField,
)

from pulp_rpm.app.constants import (
    CR_UPDATE_REFERENCE_ATTRS,
    PULP_UPDATE_COLLECTION_ATTRS,
    PULP_UPDATE_RECORD_ATTRS,
    PULP_UPDATE_REFERENCE_ATTRS,
    SKIP_TYPES
)

from pulp_rpm.app.models import (
    Addon,
    Checksum,
    DistributionTree,
    Image,
    Modulemd,
    ModulemdDefaults,
    Package,
    PackageCategory,
    PackageEnvironment,
    PackageGroup,
    PackageLangpacks,
    RepoMetadataFile,
    RpmDistribution,
    RpmRemote,
    RpmRepository,
    RpmPublication,
    UpdateRecord,
    UpdateCollection,
    Variant,
    UpdateCollectionPackage,
    UpdateReference
)
from pulp_rpm.app.shared_utils import _prepare_package
from pulp_rpm.app.schema import COPY_CONFIG_SCHEMA


class PackageSerializer(SingleArtifactContentUploadSerializer, ContentChecksumSerializer):
    """
    A Serializer for Package.

    Add serializers for the new fields defined in Package and add those fields to the Meta class
    keeping fields from the parent class as well. Provide help_text.
    """

    name = serializers.CharField(
        help_text=_("Name of the package"),
        read_only=True,
    )
    epoch = serializers.CharField(
        help_text=_("The package's epoch"),
        allow_blank=True, required=False, read_only=True,
    )
    version = serializers.CharField(
        help_text=_("The version of the package. For example, '2.8.0'"),
        read_only=True,
    )
    release = serializers.CharField(
        help_text=_("The release of a particular version of the package. e.g. '1.el7' or '3.f24'"),
        read_only=True,
    )
    arch = serializers.CharField(
        help_text=_("The target architecture for a package."
                    "For example, 'x86_64', 'i686', or 'noarch'"),
        read_only=True,
    )

    pkgId = serializers.CharField(
        help_text=_("Checksum of the package file"),
        read_only=True,
    )
    checksum_type = serializers.CharField(
        help_text=_("Type of checksum, e.g. 'sha256', 'md5'"),
        read_only=True,
    )

    summary = serializers.CharField(
        help_text=_("Short description of the packaged software"),
        allow_blank=True, required=False, read_only=True,
    )
    description = serializers.CharField(
        help_text=_("In-depth description of the packaged software"),
        allow_blank=True, required=False, read_only=True,
    )
    url = serializers.CharField(
        help_text=_("URL with more information about the packaged software"),
        allow_blank=True, required=False, read_only=True,
    )

    changelogs = serializers.JSONField(
        help_text=_("Changelogs that package contains"),
        default="[]", required=False, read_only=True,
    )
    files = serializers.JSONField(
        help_text=_("Files that package contains"),
        default="[]", required=False, read_only=True,
    )

    requires = serializers.JSONField(
        help_text=_("Capabilities the package requires"),
        default="[]", required=False, read_only=True,
    )
    provides = serializers.JSONField(
        help_text=_("Capabilities the package provides"),
        default="[]", required=False, read_only=True,
    )
    conflicts = serializers.JSONField(
        help_text=_("Capabilities the package conflicts"),
        default="[]", required=False, read_only=True,
    )
    obsoletes = serializers.JSONField(
        help_text=_("Capabilities the package obsoletes"),
        default="[]", required=False, read_only=True,
    )
    suggests = serializers.JSONField(
        help_text=_("Capabilities the package suggests"),
        default="[]", required=False, read_only=True,
    )
    enhances = serializers.JSONField(
        help_text=_("Capabilities the package enhances"),
        default="[]", required=False, read_only=True,
    )
    recommends = serializers.JSONField(
        help_text=_("Capabilities the package recommends"),
        default="[]", required=False, read_only=True,
    )
    supplements = serializers.JSONField(
        help_text=_("Capabilities the package supplements"),
        default="[]", required=False, read_only=True,
    )

    location_base = serializers.CharField(
        help_text=_("Base location of this package"),
        allow_blank=True, required=False, read_only=True,
    )
    location_href = serializers.CharField(
        help_text=_("Relative location of package to the repodata"),
        read_only=True,
    )

    rpm_buildhost = serializers.CharField(
        help_text=_("Hostname of the system that built the package"),
        allow_blank=True, required=False, read_only=True,
    )
    rpm_group = serializers.CharField(
        help_text=_("RPM group (See: http://fedoraproject.org/wiki/RPMGroups)"),
        allow_blank=True, required=False, read_only=True,
    )
    rpm_license = serializers.CharField(
        help_text=_("License term applicable to the package software (GPLv2, etc.)"),
        allow_blank=True, required=False, read_only=True,
    )
    rpm_packager = serializers.CharField(
        help_text=_("Person or persons responsible for creating the package"),
        allow_blank=True, required=False, read_only=True,
    )
    rpm_sourcerpm = serializers.CharField(
        help_text=_("Name of the source package (srpm) the package was built from"),
        allow_blank=True, required=False, read_only=True,
    )
    rpm_vendor = serializers.CharField(
        help_text=_("Name of the organization that produced the package"),
        allow_blank=True, required=False, read_only=True,
    )
    rpm_header_start = serializers.IntegerField(
        help_text=_("First byte of the header"),
        read_only=True,
    )
    rpm_header_end = serializers.IntegerField(
        help_text=_("Last byte of the header"),
        read_only=True,
    )
    is_modular = serializers.BooleanField(
        help_text=_("Flag to identify if the package is modular"),
        required=False, read_only=True,
    )

    size_archive = serializers.IntegerField(
        help_text=_("Size, in bytes, of the archive portion of the original package file"),
        read_only=True,
    )
    size_installed = serializers.IntegerField(
        help_text=_("Total size, in bytes, of every file installed by this package"),
        read_only=True,
    )
    size_package = serializers.IntegerField(
        help_text=_("Size, in bytes, of the package"),
        read_only=True,
    )

    time_build = serializers.IntegerField(
        help_text=_("Time the package was built in seconds since the epoch"),
        read_only=True,
    )
    time_file = serializers.IntegerField(
        help_text=_("The 'file' time attribute in the primary XML - "
                    "file mtime in seconds since the epoch."),
        read_only=True,
    )

    def deferred_validate(self, data):
        """
        Validate the rpm package data.

        Args:
            data (dict): Data to be validated

        Returns:
            dict: Data that has been validated

        """
        data = super().deferred_validate(data)
        # export META from rpm and prepare dict as saveable format
        try:
            new_pkg = _prepare_package(data["artifact"], data["relative_path"])
        except OSError:
            raise NotAcceptable(detail='RPM file cannot be parsed for metadata.')

        attrs = {key: new_pkg[key] for key in Package.natural_key_fields()}
        package = Package.objects.filter(**attrs)

        if package.exists():
            keywords = ('name', 'epoch', 'version', 'release', 'arch', 'checksum_type', 'pkgId')
            error_data = ", ".join(
                ["=".join(item) for item in new_pkg.items() if item[0] in keywords]
            )

            raise serializers.ValidationError(
                _(
                    "There is already a package with: {values}."
                ).format(values=error_data)
            )

        data.update(new_pkg)
        return data

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'name', 'epoch', 'version', 'release', 'arch', 'pkgId', 'checksum_type',
            'summary', 'description', 'url', 'changelogs', 'files',
            'requires', 'provides', 'conflicts', 'obsoletes',
            'suggests', 'enhances', 'recommends', 'sha256',
            'supplements', 'location_base', 'location_href',
            'rpm_buildhost', 'rpm_group', 'rpm_license',
            'rpm_packager', 'rpm_sourcerpm', 'rpm_vendor',
            'rpm_header_start', 'rpm_header_end', 'is_modular',
            'size_archive', 'size_installed', 'size_package',
            'time_build', 'time_file'
        )
        model = Package


class MinimalPackageSerializer(PackageSerializer):
    """
    A minimal serializer for RPM packages.
    """

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'name', 'epoch', 'version', 'release', 'arch', 'pkgId', 'checksum_type',
        )
        model = Package


class RpmRepositorySerializer(RepositorySerializer):
    """
    Serializer for Rpm Repositories.
    """

    metadata_signing_service = serializers.HyperlinkedRelatedField(
        help_text="A reference to an associated signing service.",
        view_name='signing-services-detail',
        queryset=AsciiArmoredDetachedSigningService.objects.all(),
        many=False,
        required=False,
        allow_null=True
    )

    class Meta:
        fields = RepositorySerializer.Meta.fields + ('metadata_signing_service',)
        model = RpmRepository


class RpmRemoteSerializer(RemoteSerializer):
    """
    A Serializer for RpmRemote.
    """

    policy = serializers.ChoiceField(
        help_text="The policy to use when downloading content. The possible values include: "
                  "'immediate', 'on_demand', and 'streamed'. 'immediate' is the default.",
        choices=Remote.POLICY_CHOICES,
        default=Remote.IMMEDIATE
    )

    class Meta:
        fields = RemoteSerializer.Meta.fields
        model = RpmRemote


class RpmPublicationSerializer(PublicationSerializer):
    """
    A Serializer for RpmPublication.
    """

    metadata_checksum_type = serializers.ChoiceField(
        help_text=_("The checksum type for metadata."),
        choices=CHECKSUM_CHOICES,
        default=CHECKSUM_TYPES.SHA256,
    )
    package_checksum_type = serializers.ChoiceField(
        help_text=_("The checksum type for packages."),
        choices=CHECKSUM_CHOICES,
        default=CHECKSUM_TYPES.SHA256,
    )

    class Meta:
        fields = PublicationSerializer.Meta.fields + (
            "metadata_checksum_type", "package_checksum_type"
        )
        model = RpmPublication


class UpdateCollectionSerializer(ModelSerializer):
    """
    A Serializer for UpdateCollection.
    """

    name = serializers.CharField(
        help_text=_("Collection name."),
        allow_blank=True,
        allow_null=True
    )

    shortname = serializers.CharField(
        help_text=_("Collection short name."),
        allow_blank=True,
        allow_null=True
    )

    packages = UpdateCollectionPackagesField(
        source='*', read_only=True,
        help_text=_("List of packages")
    )

    class Meta:
        fields = ("name", "shortname", "packages")
        model = UpdateCollection


class UpdateRecordSerializer(SingleArtifactContentUploadSerializer):
    """
    A Serializer for UpdateRecord.
    """

    id = serializers.CharField(
        help_text=_("Update id (short update name, e.g. RHEA-2013:1777)"),
        read_only=True
    )
    updated_date = serializers.CharField(
        help_text=_("Date when the update was updated (e.g. '2013-12-02 00:00:00')"),
        read_only=True
    )

    description = serializers.CharField(
        help_text=_("Update description"),
        allow_blank=True,
        read_only=True
    )
    issued_date = serializers.CharField(
        help_text=_("Date when the update was issued (e.g. '2013-12-02 00:00:00')"),
        read_only=True
    )
    fromstr = serializers.CharField(
        help_text=_("Source of the update (e.g. security@redhat.com)"),
        allow_blank=True,
        read_only=True
    )
    status = serializers.CharField(
        help_text=_("Update status ('final', ...)"),
        allow_blank=True,
        read_only=True
    )
    title = serializers.CharField(
        help_text=_("Update name"),
        allow_blank=True,
        read_only=True
    )
    summary = serializers.CharField(
        help_text=_("Short summary"),
        allow_blank=True,
        read_only=True
    )
    version = serializers.CharField(
        help_text=_("Update version (probably always an integer number)"),
        allow_blank=True,
        read_only=True
    )

    type = serializers.CharField(
        help_text=_("Update type ('enhancement', 'bugfix', ...)"),
        allow_blank=True,
        read_only=True
    )
    severity = serializers.CharField(
        help_text=_("Severity"),
        allow_blank=True,
        read_only=True
    )
    solution = serializers.CharField(
        help_text=_("Solution"),
        allow_blank=True,
        read_only=True
    )
    release = serializers.CharField(
        help_text=_("Update release"),
        allow_blank=True,
        read_only=True
    )
    rights = serializers.CharField(
        help_text=_("Copyrights"),
        allow_blank=True,
        read_only=True
    )
    pushcount = serializers.CharField(
        help_text=_("Push count"),
        allow_blank=True,
        read_only=True
    )
    reboot_suggested = serializers.BooleanField(
        help_text=_("Reboot suggested"),
        read_only=True
    )
    pkglist = UpdateCollectionSerializer(
        source='collections', read_only=True,
        many=True, help_text=_("List of packages")
    )
    references = UpdateReferenceField(
        source='pk', read_only=True,
        help_text=_("List of references")
    )

    def create(self, validated_data):
        """
        Create UpdateRecord and its subclasses from JSON file.

        Returns:
            UpdateRecord instance

        """
        references = validated_data.pop("references", [])
        pkglist = validated_data.pop("pkglist", [])
        update_collection_packages_to_save = list()
        update_references_to_save = list()
        try:
            update_record = super().create(validated_data)
        except IntegrityError:
            raise serializers.ValidationError("Advisory already exists in Pulp.")

        for collection in pkglist:
            new_coll = copy.deepcopy(collection)
            packages = new_coll.pop("packages", [])
            new_coll[PULP_UPDATE_COLLECTION_ATTRS.SHORTNAME] = new_coll.pop(
                "short", ""
            )
            coll = UpdateCollection(**new_coll)
            coll.save()
            coll.update_record.add(update_record)
            for package in packages:
                pkg = UpdateCollectionPackage(**package)
                try:
                    pkg.sum_type = createrepo_c.checksum_type(pkg.sum_type)
                except TypeError:
                    raise TypeError(f'"{pkg.sum_type}" is not supported.')
                pkg.update_collection = coll
                update_collection_packages_to_save.append(pkg)
        for reference in references:
            new_ref = dict()
            new_ref[PULP_UPDATE_REFERENCE_ATTRS.HREF] = reference.get(
                CR_UPDATE_REFERENCE_ATTRS.HREF, ""
            )
            new_ref[PULP_UPDATE_REFERENCE_ATTRS.ID] = reference.get(
                CR_UPDATE_REFERENCE_ATTRS.ID, ""
            )
            new_ref[PULP_UPDATE_REFERENCE_ATTRS.TITLE] = reference.get(
                CR_UPDATE_REFERENCE_ATTRS.TITLE, ""
            )
            new_ref[PULP_UPDATE_REFERENCE_ATTRS.TYPE] = reference.get(
                CR_UPDATE_REFERENCE_ATTRS.TYPE, ""
            )
            ref = UpdateReference(**new_ref)
            ref.update_record = update_record
            update_references_to_save.append(ref)

        if update_collection_packages_to_save:
            UpdateCollectionPackage.objects.bulk_create(update_collection_packages_to_save)
        if update_references_to_save:
            UpdateReference.objects.bulk_create(update_references_to_save)

        cr_update_record = update_record.to_createrepo_c()
        update_record.digest = hash_update_record(cr_update_record)
        update_record.save()

        return update_record

    def validate(self, data):
        """
        Read a file for a JSON data and validate a UpdateRecord data.

        Also change few fields to match Pulp internals if exists as this is usually handle by
        createrepo_c which is not used here.
        """
        update_record_data = dict()
        if 'file' in data:
            update_record_data.update(json.loads(data['file'].read()))
            update_record_data.update(data)
        elif 'artifact' in data:
            update_record_data.update(json.loads(data['artifact'].file.read()))
            update_record_data.update(data)
        else:
            raise serializers.ValidationError("Only creation with file or artifact is allowed.")

        update_record_data[PULP_UPDATE_RECORD_ATTRS.FROMSTR] = update_record_data.pop(
            'from', update_record_data.get(PULP_UPDATE_RECORD_ATTRS.FROMSTR, "")
        )
        update_record_data[PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE] = update_record_data.pop(
            'issued', update_record_data.get(PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE, "")
        )
        update_record_data[PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE] = update_record_data.pop(
            'updated', update_record_data.get(PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE, "")
        )

        if not update_record_data.get(PULP_UPDATE_RECORD_ATTRS.ID) or \
           not update_record_data.get(PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE) or \
           not update_record_data.get(PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE):
            raise serializers.ValidationError(
                "All '{}', '{}' and '{}' must be specified.".format(
                    PULP_UPDATE_RECORD_ATTRS.ID,
                    PULP_UPDATE_RECORD_ATTRS.UPDATED_DATE,
                    PULP_UPDATE_RECORD_ATTRS.ISSUED_DATE
                )
            )

        validated_data = super().validate(update_record_data)
        return validated_data

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'id', 'updated_date', 'description', 'issued_date',
            'fromstr', 'status', 'title', 'summary', 'version',
            'type', 'severity', 'solution', 'release', 'rights',
            'pushcount', 'pkglist', 'references', 'reboot_suggested'
        )
        model = UpdateRecord


class MinimalUpdateRecordSerializer(UpdateRecordSerializer):
    """
    A minimal serializer for RPM update records.
    """

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'id', 'title', 'severity', 'type'
        )
        model = UpdateRecord


class RpmDistributionSerializer(PublicationDistributionSerializer):
    """
    Serializer for RPM Distributions.
    """

    class Meta:
        fields = PublicationDistributionSerializer.Meta.fields
        model = RpmDistribution


class RpmRepositorySyncURLSerializer(RepositorySyncURLSerializer):
    """
    Serializer for RPM Sync.
    """

    skip_types = serializers.ListField(
        help_text=_("List of content types to skip during sync."),
        required=False,
        default=[],
        child=serializers.ChoiceField(
            [(skip_type, skip_type) for skip_type in SKIP_TYPES]
        )
    )
    optimize = serializers.BooleanField(
        help_text=_("Whether or not to optimize sync."),
        required=False,
        default=True
    )


class CopySerializer(serializers.Serializer):
    """
    A serializer for Content Copy API.
    """

    config = serializers.JSONField(
        help_text=_("A JSON document describing sources, destinations, and content to be copied"),
    )

    dependency_solving = serializers.BooleanField(
        help_text=_('Also copy dependencies of the content being copied.'),
        default=True
    )

    def validate(self, data):
        """
        Validate that the Serializer contains valid data.

        Set the RpmRepository based on the RepositoryVersion if only the latter is provided.
        Set the RepositoryVersion based on the RpmRepository if only the latter is provided.
        Convert the human-friendly names of the content types into what Pulp needs to query on.

        """
        super().validate(data)

        if hasattr(self, 'initial_data'):
            validate_unknown_fields(self.initial_data, self.fields)

        if 'config' in data:
            validator = Draft7Validator(COPY_CONFIG_SCHEMA)

            err = []
            for error in sorted(validator.iter_errors(data['config']), key=str):
                err.append(error.message)
            if err:
                raise serializers.ValidationError(
                    _("Provided copy criteria is invalid:'{}'".format(err))
                )

        return data


class PackageGroupSerializer(NoArtifactContentSerializer):
    """
    PackageGroup serializer.
    """

    id = serializers.CharField(
        help_text=_("PackageGroup id."),
    )
    default = serializers.BooleanField(
        help_text=_("PackageGroup default."),
        required=False
    )
    user_visible = serializers.BooleanField(
        help_text=_("PackageGroup user visibility."),
        required=False
    )
    display_order = serializers.IntegerField(
        help_text=_("PackageGroup display order."),
        allow_null=True
    )
    name = serializers.CharField(
        help_text=_("PackageGroup name."),
        allow_blank=True
    )
    description = serializers.CharField(
        help_text=_("PackageGroup description."),
        allow_blank=True
    )
    packages = serializers.JSONField(
        help_text=_("PackageGroup package list."),
        allow_null=True
    )
    biarch_only = serializers.BooleanField(
        help_text=_("PackageGroup biarch only."),
        required=False
    )
    desc_by_lang = serializers.JSONField(
        help_text=_("PackageGroup description by language."),
        allow_null=True
    )
    name_by_lang = serializers.JSONField(
        help_text=_("PackageGroup name by language."),
        allow_null=True
    )
    digest = serializers.CharField(
        help_text=_("PackageGroup digest."),
    )
    related_packages = RelatedField(
        help_text=_("Packages related to this PackageGroup."),
        allow_null=True,
        required=False,
        queryset=Package.objects.all(),
        many=True,
        view_name='content-rpm/packages-detail'
    )

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            'id', 'default', 'user_visible', 'display_order',
            'name', 'description', 'packages', 'biarch_only',
            'desc_by_lang', 'name_by_lang', 'digest', 'related_packages'
        )
        model = PackageGroup


class PackageCategorySerializer(NoArtifactContentSerializer):
    """
    PackageCategory serializer.
    """

    id = serializers.CharField(
        help_text=_("Category id."),
    )
    name = serializers.CharField(
        help_text=_("Category name."),
        allow_blank=True
    )
    description = serializers.CharField(
        help_text=_("Category description."),
        allow_blank=True
    )
    display_order = serializers.IntegerField(
        help_text=_("Category display order."),
        allow_null=True
    )
    group_ids = serializers.JSONField(
        help_text=_("Category group list."),
        allow_null=True
    )
    desc_by_lang = serializers.JSONField(
        help_text=_("Category description by language."),
        allow_null=True
    )
    name_by_lang = serializers.JSONField(
        help_text=_("Category name by language."),
        allow_null=True
    )
    digest = serializers.CharField(
        help_text=_("Category digest."),
    )
    packagegroups = RelatedField(
        help_text=_("PackageGroups related to this category."),
        allow_null=True,
        required=False,
        queryset=PackageGroup.objects.all(),
        many=True,
        view_name='content-rpm/packagegroups-detail'
    )

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            'id', 'name', 'description', 'display_order',
            'group_ids', 'desc_by_lang', 'name_by_lang', 'digest',
            'packagegroups'
        )
        model = PackageCategory


class PackageEnvironmentSerializer(NoArtifactContentSerializer):
    """
    PackageEnvironment serializer.
    """

    id = serializers.CharField(
        help_text=_("Environment id."),
    )
    name = serializers.CharField(
        help_text=_("Environment name."),
        allow_blank=True
    )
    description = serializers.CharField(
        help_text=_("Environment description."),
        allow_blank=True
    )
    display_order = serializers.IntegerField(
        help_text=_("Environment display order."),
        allow_null=True
    )
    group_ids = serializers.JSONField(
        help_text=_("Environment group list."),
        allow_null=True
    )
    option_ids = serializers.JSONField(
        help_text=_("Environment option ids"),
        allow_null=True
    )
    desc_by_lang = serializers.JSONField(
        help_text=_("Environment description by language."),
        allow_null=True
    )
    name_by_lang = serializers.JSONField(
        help_text=_("Environment name by language."),
        allow_null=True
    )
    digest = serializers.CharField(
        help_text=_("Environment digest.")
    )
    packagegroups = RelatedField(
        help_text=_("Groups related to this Environment."),
        allow_null=True,
        required=False,
        queryset=PackageGroup.objects.all(),
        many=True,
        view_name='content-rpm/packagegroups-detail'
    )
    optionalgroups = RelatedField(
        help_text=_("Groups optionally related to this Environment."),
        allow_null=True,
        required=False,
        queryset=PackageGroup.objects.all(),
        many=True,
        view_name='content-rpm/packagegroups-detail'
    )

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            'id', 'name', 'description', 'display_order',
            'group_ids', 'option_ids', 'desc_by_lang', 'name_by_lang',
            'digest', 'packagegroups', 'optionalgroups'
        )
        model = PackageEnvironment


class PackageLangpacksSerializer(NoArtifactContentSerializer):
    """
    PackageLangpacks serializer.
    """

    matches = serializers.JSONField(
        help_text=_("Langpacks matches."),
        allow_null=True
    )
    digest = serializers.CharField(
        help_text=_("Langpacks digest."),
        allow_null=True
    )

    class Meta:
        fields = NoArtifactContentSerializer.Meta.fields + (
            'matches', 'digest'
        )
        model = PackageLangpacks


class ModulemdSerializer(SingleArtifactContentUploadSerializer, ContentChecksumSerializer):
    """
    Modulemd serializer.
    """

    name = serializers.CharField(
        help_text=_("Modulemd name."),
    )
    stream = serializers.CharField(
        help_text=_("Stream name."),
    )
    version = serializers.CharField(
        help_text=_("Modulemd version."),
    )
    context = serializers.CharField(
        help_text=_("Modulemd context."),
    )
    arch = serializers.CharField(
        help_text=_("Modulemd architecture."),
    )
    artifacts = serializers.JSONField(
        help_text=_("Modulemd artifacts."),
        allow_null=True
    )
    dependencies = serializers.JSONField(
        help_text=_("Modulemd dependencies."),
        allow_null=True
    )
    # TODO: The performance of this is not great, there's a noticable difference in response
    # time before/after. Since this will only return Package content hrefs, we might benefit
    # from creating a specialized version of this Field that can skip some of the work.
    packages = DetailRelatedField(
        help_text=_("Modulemd artifacts' packages."),
        allow_null=True,
        required=False,
        queryset=Package.objects.all(),
        view_name="content-rpm/packages-detail",
        many=True
    )

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'name', 'stream', 'version', 'context', 'arch',
            'artifacts', 'dependencies', 'packages', 'sha256'
        )
        model = Modulemd


class ModulemdDefaultsSerializer(SingleArtifactContentUploadSerializer, ContentChecksumSerializer):
    """
    ModulemdDefaults serializer.
    """

    module = serializers.CharField(
        help_text=_("Modulemd name.")
    )
    stream = serializers.CharField(
        help_text=_("Modulemd default stream.")
    )
    profiles = serializers.JSONField(
        help_text=_("Default profiles for modulemd streams.")
    )

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'module', 'stream', 'profiles', 'sha256'
        )
        model = ModulemdDefaults


class AddonSerializer(serializers.ModelSerializer):
    """
    Addon serializer.
    """

    addon_id = serializers.CharField(help_text=_("Addon id."))
    uid = serializers.CharField(help_text=_("Addon uid."))
    name = serializers.CharField(help_text=_("Addon name."))
    type = serializers.CharField(help_text=_("Addon type."))
    packages = serializers.CharField(
        help_text=_("Relative path to directory with binary RPMs.")
    )

    class Meta:
        model = Addon
        fields = (
            "addon_id", "uid", "name", "type", "packages"
        )


class ChecksumSerializer(serializers.ModelSerializer):
    """
    Checksum serializer.
    """

    path = serializers.CharField(help_text=_("File path."))
    checksum = serializers.CharField(help_text=_("Checksum for the file."))

    class Meta:
        model = Checksum
        fields = ("path", "checksum")


class ImageSerializer(serializers.ModelSerializer):
    """
    Image serializer.
    """

    name = serializers.CharField(help_text=_("File name."))
    path = serializers.CharField(help_text=_("File path."))
    platforms = serializers.CharField(help_text=_("Compatible platforms."))
    artifact = ArtifactSerializer()

    class Meta:
        model = Image
        fields = ("name", "path", "platforms", "artifact")


class VariantSerializer(serializers.ModelSerializer):
    """
    Variant serializer.
    """

    variant_id = serializers.CharField(help_text=_("Variant id."))
    uid = serializers.CharField(help_text=_("Variant uid."))
    name = serializers.CharField(help_text=_("Variant name."))
    type = serializers.CharField(help_text=_("Variant type."))
    packages = serializers.CharField(
        help_text=_("Relative path to directory with binary RPMs.")
    )
    source_packages = serializers.CharField(
        help_text=_("Relative path to directory with source RPMs.")
    )
    source_repository = serializers.CharField(
        help_text=_("Relative path to YUM repository with source RPMs.")
    )
    debug_packages = serializers.CharField(
        help_text=_("Relative path to directory with debug RPMs.")
    )
    debug_repository = serializers.CharField(
        help_text=_("Relative path to YUM repository with debug RPMs.")
    )
    identity = serializers.CharField(
        help_text=_("Relative path to a pem file that identifies a product.")
    )

    class Meta:
        model = Variant
        fields = (
            "variant_id", "uid", "name", "type", "packages",
            "source_packages", "source_repository", "debug_packages",
            "debug_repository", "identity"
        )


class DistributionTreeSerializer(MultipleArtifactContentSerializer):
    """
    DistributionTree serializer.
    """

    header_version = serializers.CharField(
        help_text=_("Header Version.")
    )
    release_name = serializers.CharField(
        help_text=_("Release name.")
    )
    release_short = serializers.CharField(
        help_text=_("Release short name.")
    )
    release_version = serializers.CharField(
        help_text=_("Release version.")
    )
    release_is_layered = serializers.BooleanField(
        help_text=_("Typically False for an operating system, True otherwise.")
    )

    base_product_name = serializers.CharField(
        help_text=_("Base Product name."), allow_null=True
    )
    base_product_short = serializers.CharField(
        help_text=_("Base Product short name."), allow_null=True
    )
    base_product_version = serializers.CharField(
        help_text=_("Base Product version."), allow_null=True
    )

    arch = serializers.CharField(help_text=_("Tree architecturerch."))
    build_timestamp = serializers.FloatField(help_text=_("Tree build time timestamp."))

    instimage = serializers.CharField(
        help_text=_("Relative path to Anaconda instimage."), allow_null=True
    )
    mainimage = serializers.CharField(
        help_text=_("Relative path to Anaconda stage2 image."), allow_null=True
    )

    discnum = serializers.IntegerField(
        help_text=_("Disc number."), allow_null=True
    )
    totaldiscs = serializers.IntegerField(
        help_text=_("Number of discs in media set."), allow_null=True
    )

    addons = AddonSerializer(many=True)

    checksums = ChecksumSerializer(many=True)

    images = ImageSerializer(many=True)

    variants = VariantSerializer(many=True)

    class Meta:
        model = DistributionTree
        fields = (
            "pulp_href", "header_version", "release_name", "release_short", "release_version",
            "release_is_layered", "base_product_name", "base_product_short",
            "base_product_version", "arch", "build_timestamp", "instimage", "mainimage",
            "discnum", "totaldiscs", "addons", "checksums", "images", "variants"
        )


class RepoMetadataFileSerializer(SingleArtifactContentUploadSerializer, ContentChecksumSerializer):
    """
    RepoMetadataFile serializer.
    """

    data_type = serializers.CharField(
        help_text=_("Metadata type.")
    )
    checksum_type = serializers.CharField(
        help_text=_("Checksum type for the file.")
    )
    checksum = serializers.CharField(
        help_text=_("Checksum for the file.")
    )

    class Meta:
        fields = SingleArtifactContentUploadSerializer.Meta.fields + (
            'data_type', 'checksum_type', 'checksum', 'sha256'
        )
        model = RepoMetadataFile



# Generated by Django 2.2.11 on 2020-03-23 15:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0005_optimize_sync'),
    ]

    operations = [
        migrations.AddField(
            model_name='updatecollectionpackage',
            name='relogin_suggested',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='updatecollectionpackage',
            name='restart_suggested',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name='updatecollection',
            name='name',
            field=models.TextField(null=True),
        ),
        migrations.AlterField(
            model_name='updatecollection',
            name='shortname',
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 2.2.12 on 2020-05-06 15:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0008_advisory_pkg_sumtype_as_int'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rpmrepository',
            name='last_sync_revision_number',
            field=models.CharField(max_length=20, null=True),
        ),
    ]



# coding=utf-8
"""Tests that publish rpm plugin repositories."""
import unittest
from datetime import datetime
from urllib.parse import urljoin

from pulp_smash import api, config
from pulp_smash.pulp3.utils import (
    delete_orphans,
    gen_repo,
    get_added_content_summary,
    get_content,
    get_content_summary,
)

from pulp_rpm.tests.functional.constants import (
    RPM_KICKSTART_CONTENT_NAME,
    RPM_PACKAGE_CONTENT_NAME,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_PUBLICATION_PATH,
    RPM_REMOTE_PATH,
    RPM_REPO_PATH,
    CENTOS6_URL,
    CENTOS7_URL,
    CENTOS8_APPSTREAM_URL,
    CENTOS8_BASEOS_URL,
    CENTOS8_KICKSTART_APP_URL,
    CENTOS8_KICKSTART_BASEOS_URL,
    EPEL7_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_remote
from pulp_rpm.tests.functional.utils import set_up_module as setUpModule  # noqa:F401


class PublishTestCase(unittest.TestCase):
    """Publish repositories with the rpm plugin."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)

        delete_orphans(cls.cfg)

    def parse_date_from_string(self, s, parse_format='%Y-%m-%dT%H:%M:%S.%fZ'):
        """Parse string to datetime object.

        :param s: str like '2018-11-18T21:03:32.493697Z'
        :param parse_format: str defaults to %Y-%m-%dT%H:%M:%S.%fZ
        :return: datetime.datetime
        """
        return datetime.strptime(s, parse_format)

    def rpm_publish(self, url=RPM_KICKSTART_FIXTURE_URL, policy='on_demand'):
        """Publish repositories with the rpm plugin.

        This test targets the following issue:
        `Pulp #5630 <https://pulp.plan.io/issues/5630>`_.

        In order to sync a repository a remote has to be associated within
        this repository. When a repository is created this version field is set
        as None. After a sync the repository version is updated.

        Do the following:

        1. Create a repository and a remote.
        2. Assert that repository version is None.
        3. Sync the remote.
        4. Assert that repository version is not None.
        5. Assert that distribution_tree units were added and are present in the repo.
        6. Publish
        """
        delete_orphans(self.cfg)
        repo = self.client.post(RPM_REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo['pulp_href'])

        # Create a remote with the standard test fixture url.
        body = gen_rpm_remote(url=url, policy=policy)
        remote = self.client.post(RPM_REMOTE_PATH, body)
        self.addCleanup(self.client.delete, remote['pulp_href'])

        # Sync the repository.
        self.assertEqual(repo["latest_version_href"], f"{repo['pulp_href']}versions/0/")
        data = {"remote": remote["pulp_href"]}
        response = self.client.using_handler(api.json_handler).post(
            urljoin(repo["pulp_href"], "sync/"), data
        )
        sync_task = self.client.get(response["task"])
        created_at = self.parse_date_from_string(sync_task["pulp_created"])
        started_at = self.parse_date_from_string(sync_task["started_at"])
        finished_at = self.parse_date_from_string(sync_task["finished_at"])
        task_duration = finished_at - started_at
        waiting_time = started_at - created_at
        print("\n->     Sync => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(),
            service=task_duration.total_seconds()
        ))

        repo = self.client.get(repo['pulp_href'])
        for kickstart_content in get_content(repo)[RPM_KICKSTART_CONTENT_NAME]:
            self.addCleanup(self.client.delete, kickstart_content['pulp_href'])

        # Check that we have the correct content counts.
        self.assertIsNotNone(repo['latest_version_href'])

        self.assertIn(
            RPM_PACKAGE_CONTENT_NAME,
            get_content_summary(repo).keys(),
        )
        self.assertIn(
            RPM_PACKAGE_CONTENT_NAME,
            get_added_content_summary(repo).keys(),
        )

        repo = self.client.get(repo['pulp_href'])

        # Publishing
        body = {"repository": repo["pulp_href"]}
        response = self.client.using_handler(api.json_handler).post(RPM_PUBLICATION_PATH, body)
        publish_task = self.client.get(response["task"])
        created_at = self.parse_date_from_string(publish_task["pulp_created"])
        started_at = self.parse_date_from_string(publish_task["started_at"])
        finished_at = self.parse_date_from_string(publish_task["finished_at"])
        task_duration = finished_at - started_at
        waiting_time = started_at - created_at
        print("\n->     Publish => Waiting time (s): {wait} | Service time (s): {service}".format(
            wait=waiting_time.total_seconds(),
            service=task_duration.total_seconds()
        ))

    def test_epel7(self):
        """Publish EPEL 7."""
        self.rpm_publish(url=EPEL7_URL)

    def test_centos6(self):
        """Publish CentOS 6."""
        self.rpm_publish(url=CENTOS6_URL)

    def test_centos7(self):
        """Publish CentOS 7."""
        self.rpm_publish(url=CENTOS7_URL)

    def test_centos8_baseos(self):
        """Publish CentOS 8 BaseOS."""
        self.rpm_publish(url=CENTOS8_BASEOS_URL)

    def test_centos8_appstream(self):
        """Publish CentOS 8 AppStream."""
        self.rpm_publish(url=CENTOS8_APPSTREAM_URL)

    def test_centos8_kickstart_baseos(self):
        """Kickstart Publish CentOS 8 BaseOS."""
        self.rpm_publish(url=CENTOS8_KICKSTART_BASEOS_URL)

    def test_centos8_kickstart_appstream(self):
        """Kickstart Publish CentOS 8 AppStream."""
        self.rpm_publish(url=CENTOS8_KICKSTART_APP_URL)



import os
import tempfile

from pulpcore.plugin.models import Artifact
from pulp_rpm.app.models import Modulemd, Package
from pulp_rpm.app.constants import PULP_MODULE_ATTR, PULP_MODULEDEFAULTS_ATTR

import gi
gi.require_version('Modulemd', '2.0')
from gi.repository import Modulemd as mmdlib  # noqa: E402


def resolve_module_packages(version, previous_version):
    """
    Decide which packages to add/remove based on modular data.

    Args:
        version (pulpcore.app.models.RepositoryVersion): current incomplete repository version
        previous_version (pulpcore.app.models.RepositoryVersion) :  previous version of the same
                                                                    repository to compare to

    """
    def modules_packages(modules):
        packages = set()
        for module in modules:
            packages.update(module.packages.all())
        return packages

    modulemd_pulp_type = Modulemd.get_pulp_type()
    current_modules = Modulemd.objects \
        .filter(pk__in=version.content.filter(pulp_type=modulemd_pulp_type))
    current_module_packages = modules_packages(current_modules)

    if previous_version:
        previous_modules = Modulemd.objects \
            .filter(pk__in=previous_version.content.filter(pulp_type=modulemd_pulp_type))
        added_modules = current_modules.difference(previous_modules)
        removed_modules = previous_modules.difference(current_modules)
        removed_module_packages = modules_packages(removed_modules)
        packages_to_remove = removed_module_packages.difference(current_module_packages)
        version.remove_content(Package.objects.filter(pk__in=packages_to_remove))
    else:
        added_modules = current_modules

    added_module_packages = modules_packages(added_modules)
    packages_to_add = added_module_packages.difference(current_module_packages)
    version.add_content(Package.objects.filter(pk__in=packages_to_add))


def _create_snippet(snippet_string):
    """
    Create snippet of modulemd[-defaults] as artifact.

    Args:
        snippet_string (string):
            Snippet with modulemd[-defaults] yaml

    Returns:
        Snippet as unsaved Artifact object

    """
    tmp_file = tempfile.NamedTemporaryFile(dir=os.getcwd(), delete=False)
    with open(tmp_file.name, 'w') as snippet:
        snippet.write(snippet_string)
    return Artifact.init_and_validate(tmp_file.name)


def parse_modulemd(module_names, module_index):
    """
    Get modulemd NSVCA, artifacts, dependencies.

    Args:
        module_names (list):
            list of modulemd names
        module_index (mmdlib.ModuleIndex):
            libmodulemd index object

    """
    ret = list()
    for module in module_names:
        for stream in module_index.get_module(module).get_all_streams():
            modulemd = dict()
            modulemd[PULP_MODULE_ATTR.NAME] = stream.props.module_name
            modulemd[PULP_MODULE_ATTR.STREAM] = stream.props.stream_name
            modulemd[PULP_MODULE_ATTR.VERSION] = stream.props.version
            modulemd[PULP_MODULE_ATTR.CONTEXT] = stream.props.context
            modulemd[PULP_MODULE_ATTR.ARCH] = stream.props.arch
            modulemd[PULP_MODULE_ATTR.ARTIFACTS] = stream.get_rpm_artifacts()

            dependencies_list = stream.get_dependencies()
            dependencies = list()
            for dep in dependencies_list:
                depmodule_list = dep.get_runtime_modules()
                platform_deps = dict()
                for depmod in depmodule_list:
                    platform_deps[depmod] = dep.get_runtime_streams(depmod)
                dependencies.append(platform_deps)
            modulemd[PULP_MODULE_ATTR.DEPENDENCIES] = dependencies
            # create yaml snippet for this modulemd stream
            temp_index = mmdlib.ModuleIndex.new()
            temp_index.add_module_stream(stream)
            artifact = _create_snippet(temp_index.dump_to_string())
            modulemd["artifact"] = artifact
            ret.append(modulemd)
    return ret


def parse_defaults(module_index):
    """
    Get modulemd_defaults.

    Args:
        module_index (mmdlib.ModuleIndex):
            libmodulemd index object

    Returns:
        list of modulemd_defaults as dict

    """
    ret = list()
    modulemd_defaults = module_index.get_default_streams().keys()
    for module in modulemd_defaults:
        modulemd = module_index.get_module(module)
        defaults = modulemd.get_defaults()
        if defaults:
            default_stream = defaults.get_default_stream()
            default_profile = defaults.get_default_profiles_for_stream(default_stream)
            # create modulemd-default snippet
            temp_index = mmdlib.ModuleIndex.new()
            temp_index.add_defaults(defaults)
            artifact = _create_snippet(temp_index.dump_to_string())
            ret.append({
                PULP_MODULEDEFAULTS_ATTR.MODULE: modulemd.get_module_name(),
                PULP_MODULEDEFAULTS_ATTR.STREAM: default_stream,
                PULP_MODULEDEFAULTS_ATTR.PROFILES: default_profile,
                PULP_MODULEDEFAULTS_ATTR.DIGEST: artifact.sha256,
                'artifact': artifact
            })
    return ret



from logging import getLogger

from django.db import models

from pulpcore.plugin.models import (
    BaseModel,
    Content,
    ContentArtifact,
    Repository,
)

log = getLogger(__name__)


class DistributionTree(Content):
    """
    Model for an RPM distribution tree (also sometimes referenced as an installable tree).

    A distribution tree is described by a file in root of an RPM repository named either
    "treeinfo" or ".treeinfo". This INI file is used by system installers to boot from a URL.
    It describes the operating system or product contained in the distribution tree and
    where the bootable media is located for various platforms (where platform means
    'x86_64', 'xen', or similar).

    The description of the "treeinfo" format is included below, originally take from
    https://release-engineering.github.io/productmd/treeinfo-1.0.html

    Fields:
        header_version (Text):
            Metadata version
        release_name (Text):
            Release name
        release_short (Text):
            Release short name
        release_version (Text):
            Release version
        release_type (Text):
            Release type
        release_is_layered (Bool):
            Typically False for an operating system, True otherwise
        base_product_name (Text):
            Base product name
        base_product_short (Text):
            Base product short name
        base_product_version (Text):
            Base product *major* version
        base_product_type (Text):
            Base product release type
        arch (Text):
            Tree architecture
        build_timestamp (Float):
            Tree build time timestamp
        instimage (Text):
            Relative path to Anaconda instimage
        mainimage (Text):
            Relative path to Anaconda stage2 image
        discnum (Integer):
            Disc number
        totaldiscs (Integer):
            Number of discs in media set

    """

    TYPE = 'distribution_tree'

    header_version = models.CharField(max_length=10)

    release_name = models.CharField(max_length=50)
    release_short = models.CharField(max_length=20)
    release_version = models.CharField(max_length=10)
    release_is_layered = models.BooleanField(default=False)

    base_product_name = models.CharField(max_length=50, null=True)
    base_product_short = models.CharField(max_length=20, null=True)
    base_product_version = models.CharField(max_length=10, null=True)

    # tree
    arch = models.CharField(max_length=30)
    build_timestamp = models.FloatField()

    # stage2
    instimage = models.CharField(max_length=50, null=True)
    mainimage = models.CharField(max_length=50, null=True)

    # media
    discnum = models.IntegerField(null=True)
    totaldiscs = models.IntegerField(null=True)

    repo_key_fields = (
        "header_version",
        "release_name",
        "release_short",
        "release_version",
        "arch",
        "build_timestamp",
    )

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
        unique_together = (
            "header_version",
            "release_name",
            "release_short",
            "release_version",
            "arch",
            "build_timestamp",
        )


class Checksum(BaseModel):
    """
    Distribution Tree Checksum.

    Checksums of selected files in a tree.

    Fields:
        path (Text):
            File path
        checksum (Text):
            Checksum value for the file

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree

    """

    path = models.CharField(max_length=128)
    checksum = models.CharField(max_length=128, null=True)
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name='checksums'
    )

    class Meta:
        unique_together = (
            "path",
            "checksum",
            "distribution_tree",
        )


class Image(BaseModel):
    """
    Distribution Tree Image.

    Images compatible with particular platform.

    Fields:
        name (Text):
            File name
        path (Text):
            File path
        platforms (Text):
            Compatible platforms

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree

    """

    name = models.CharField(max_length=20)
    path = models.CharField(max_length=128)
    platforms = models.CharField(max_length=20)
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name='images'
    )

    @property
    def artifact(self):
        """
        Returns artifact object.
        """
        content_artifact = ContentArtifact.objects.filter(
            content=self.distribution_tree,
            relative_path=self.path,
        ).first()

        artifact = content_artifact.artifact if content_artifact else None

        return artifact

    class Meta:
        unique_together = (
            "name",
            "path",
            "platforms",
            "distribution_tree",
        )


class Addon(BaseModel):
    """
    Distribution Tree Addon.

    Kickstart functionality expansion.

    Fields:
        addon_id (Text):
            Addon id
        uid (Text):
            Addon uid
        name (Text):
            Addon name
        type (Text):
            Addon type
        packages (Text):
            Relative path to directory with binary RPMs

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree
        repository (models.ForeignKey): The associated Repository

    """

    addon_id = models.CharField(max_length=50)
    uid = models.CharField(max_length=50)
    name = models.CharField(max_length=50)
    type = models.CharField(max_length=20)
    packages = models.CharField(max_length=50)
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name='addons'
    )
    repository = models.ForeignKey(
        Repository, on_delete=models.PROTECT, related_name='addons'
    )

    class Meta:
        unique_together = (
            "addon_id",
            "uid",
            "name",
            "type",
            "packages",
            "distribution_tree",
        )


class Variant(BaseModel):
    """
    Distribution Tree Variant.

    Fields:
        variant_id (Text):
            Variant id
        uid (Text):
            Variant uid
        name (Text):
            Variant name
        type (Text):
            Variant type
        packages (Text):
            Relative path to directory with binary RPMs
        source_packages (Text):
            Relative path to directory with source RPMs
        source_repository (Text):
            Relative path to YUM repository with source RPMs
        debug_packages (Text):
            Relative path to directory with debug RPMs
        debug_repository (Text):
            Relative path to YUM repository with debug RPMs
        identity (Text):
            Relative path to a pem file that identifies a product

    Relations:

        distribution_tree (models.ForeignKey): The associated DistributionTree
        repository (models.ForeignKey): The associated Repository

    """

    variant_id = models.CharField(max_length=50)
    uid = models.CharField(max_length=50)
    name = models.CharField(max_length=50)
    type = models.CharField(max_length=20)
    packages = models.CharField(max_length=50)
    source_packages = models.CharField(max_length=50, null=True)
    source_repository = models.CharField(max_length=50, null=True)
    debug_packages = models.CharField(max_length=50, null=True)
    debug_repository = models.CharField(max_length=50, null=True)
    identity = models.CharField(max_length=50, null=True)
    distribution_tree = models.ForeignKey(
        DistributionTree, on_delete=models.CASCADE, related_name='variants'
    )
    repository = models.ForeignKey(
        Repository, on_delete=models.PROTECT, related_name='+'
    )

    class Meta:
        unique_together = (
            "variant_id",
            "uid",
            "name",
            "type",
            "packages",
            "distribution_tree",
        )



from types import SimpleNamespace

CHECKSUM_TYPES = SimpleNamespace(
    UNKNOWN='unknown',
    MD5='md5',
    SHA='sha1',  # compatibility nickname from original createrepo
    SHA1='sha1',
    SHA224='sha224',
    SHA256='sha256',
    SHA384='sha384',
    SHA512='sha512'
)

# The same as above, but in a format that choice fields can use
CHECKSUM_CHOICES = (
    (CHECKSUM_TYPES.UNKNOWN, CHECKSUM_TYPES.UNKNOWN),
    (CHECKSUM_TYPES.MD5, CHECKSUM_TYPES.MD5),
    (CHECKSUM_TYPES.SHA, CHECKSUM_TYPES.SHA),
    (CHECKSUM_TYPES.SHA1, CHECKSUM_TYPES.SHA1),
    (CHECKSUM_TYPES.SHA224, CHECKSUM_TYPES.SHA224),
    (CHECKSUM_TYPES.SHA256, CHECKSUM_TYPES.SHA256),
    (CHECKSUM_TYPES.SHA384, CHECKSUM_TYPES.SHA384),
    (CHECKSUM_TYPES.SHA512, CHECKSUM_TYPES.SHA512)
)

CR_PACKAGE_ATTRS = SimpleNamespace(
    ARCH='arch',
    CHANGELOGS='changelogs',
    CHECKSUM_TYPE='checksum_type',
    CONFLICTS='conflicts',
    DESCRIPTION='description',
    ENHANCES='enhances',
    EPOCH='epoch',
    FILES='files',
    LOCATION_BASE='location_base',
    LOCATION_HREF='location_href',
    NAME='name',
    OBSOLETES='obsoletes',
    PKGID='pkgId',
    PROVIDES='provides',
    RECOMMENDS='recommends',
    RELEASE='release',
    REQUIRES='requires',
    RPM_BUILDHOST='rpm_buildhost',
    RPM_GROUP='rpm_group',
    RPM_HEADER_END='rpm_header_end',
    RPM_HEADER_START='rpm_header_start',
    RPM_LICENSE='rpm_license',
    RPM_PACKAGER='rpm_packager',
    RPM_SOURCERPM='rpm_sourcerpm',
    RPM_VENDOR='rpm_vendor',
    SIZE_ARCHIVE='size_archive',
    SIZE_INSTALLED='size_installed',
    SIZE_PACKAGE='size_package',
    SUGGESTS='suggests',
    SUMMARY='summary',
    SUPPLEMENTS='supplements',
    TIME_BUILD='time_build',
    TIME_FILE='time_file',
    URL='url',
    VERSION='version'
)

PACKAGE_REPODATA = ['primary', 'filelists', 'other']
PACKAGE_DB_REPODATA = ['primary_db', 'filelists_db', 'other_db']
UPDATE_REPODATA = ['updateinfo']
MODULAR_REPODATA = ['modules']
COMPS_REPODATA = ['group']
SKIP_REPODATA = ['group_gz', 'prestodelta']
SKIP_TYPES = ['srpm']

CR_UPDATE_RECORD_ATTRS = SimpleNamespace(
    ID='id',
    UPDATED_DATE='updated_date',
    DESCRIPTION='description',
    ISSUED_DATE='issued_date',
    FROMSTR='fromstr',
    STATUS='status',
    TITLE='title',
    SUMMARY='summary',
    VERSION='version',
    TYPE='type',
    SEVERITY='severity',
    SOLUTION='solution',
    RELEASE='release',
    RIGHTS='rights',
    PUSHCOUNT='pushcount',
    REBOOT_SUGGESTED='reboot_suggested'
)

CR_UPDATE_COLLECTION_ATTRS = SimpleNamespace(
    NAME='name',
    SHORTNAME='shortname',
    MODULE='module'
)

CR_UPDATE_COLLECTION_PACKAGE_ATTRS = SimpleNamespace(
    ARCH='arch',
    EPOCH='epoch',
    FILENAME='filename',
    NAME='name',
    REBOOT_SUGGESTED='reboot_suggested',
    RELOGIN_SUGGESTED='relogin_suggested',
    RESTART_SUGGESTED='restart_suggested',
    RELEASE='release',
    SRC='src',
    SUM='sum',
    SUM_TYPE='sum_type',
    VERSION='version'
)

CR_UPDATE_REFERENCE_ATTRS = SimpleNamespace(
    HREF='href',
    ID='id',
    TITLE='title',
    TYPE='type'
)

CR_UPDATE_COLLECTION_ATTRS_MODULE = SimpleNamespace(
    NAME='name',
    STREAM='stream',
    VERSION='version',
    CONTEXT='context',
    ARCH='arch'
)

PULP_UPDATE_REFERENCE_ATTRS = SimpleNamespace(
    HREF='href',
    ID='ref_id',
    TITLE='title',
    TYPE='ref_type'
)

PULP_PACKAGE_ATTRS = CR_PACKAGE_ATTRS
PULP_UPDATE_RECORD_ATTRS = CR_UPDATE_RECORD_ATTRS
PULP_UPDATE_COLLECTION_ATTRS = CR_UPDATE_COLLECTION_ATTRS
PULP_UPDATE_COLLECTION_PACKAGE_ATTRS = CR_UPDATE_COLLECTION_PACKAGE_ATTRS

PULP_UPDATE_COLLECTION_ATTRS_MODULE = CR_UPDATE_COLLECTION_ATTRS_MODULE

MODULEMD_MODULE_ATTR = SimpleNamespace(
    ARCH='arch',
    ARTIFACTS='artifacts',
    CONTEXT='context',
    NAME='name',
    STREAM='stream',
    VERSION='version',
    DEPENDENCIES='dependencies'
)

MODULEMD_MODULEDEFAULTS_ATTR = SimpleNamespace(
    MODULE='module',
    STREAM='stream',
    PROFILES='profiles'
)

PULP_MODULEDEFAULTS_ATTR = SimpleNamespace(
    MODULE='module',
    STREAM='stream',
    PROFILES='profiles',
    DIGEST='digest'
)

PULP_MODULE_ATTR = MODULEMD_MODULE_ATTR

LIBCOMPS_GROUP_ATTRS = SimpleNamespace(
    ID='id',
    DEFAULT='default',
    USER_VISIBLE='uservisible',
    DISPLAY_ORDER='display_order',
    NAME='name',
    DESCRIPTION='desc',
    PACKAGES='packages',
    BIARCH_ONLY='biarchonly',
    DESC_BY_LANG='desc_by_lang',
    NAME_BY_LANG='name_by_lang'
)

LIBCOMPS_CATEGORY_ATTRS = SimpleNamespace(
    ID='id',
    NAME='name',
    DESCRIPTION='desc',
    DISPLAY_ORDER='display_order',
    GROUP_IDS='group_ids',
    DESC_BY_LANG='desc_by_lang',
    NAME_BY_LANG='name_by_lang'
)

LIBCOMPS_ENVIRONMENT_ATTRS = SimpleNamespace(
    ID='id',
    NAME='name',
    DESCRIPTION='desc',
    DISPLAY_ORDER='display_order',
    GROUP_IDS='group_ids',
    OPTION_IDS='option_ids',
    DESC_BY_LANG='desc_by_lang',
    NAME_BY_LANG='name_by_lang'
)

PULP_LANGPACKS_ATTRS = SimpleNamespace(
    MATCHES='matches'
)

PULP_GROUP_ATTRS = SimpleNamespace(
    ID='id',
    DEFAULT='default',
    USER_VISIBLE='user_visible',
    DISPLAY_ORDER='display_order',
    NAME='name',
    DESCRIPTION='description',
    PACKAGES='packages',
    BIARCH_ONLY='biarch_only',
    DESC_BY_LANG='desc_by_lang',
    NAME_BY_LANG='name_by_lang'
)

PULP_CATEGORY_ATTRS = SimpleNamespace(
    ID='id',
    NAME='name',
    DESCRIPTION='description',
    DISPLAY_ORDER='display_order',
    GROUP_IDS='group_ids',
    DESC_BY_LANG='desc_by_lang',
    NAME_BY_LANG='name_by_lang'
)

PULP_ENVIRONMENT_ATTRS = SimpleNamespace(
    ID='id',
    NAME='name',
    DESCRIPTION='description',
    DISPLAY_ORDER='display_order',
    GROUP_IDS='group_ids',
    OPTION_IDS='option_ids',
    DESC_BY_LANG='desc_by_lang',
    NAME_BY_LANG='name_by_lang'
)

PACKAGES_DIRECTORY = "Packages"

# Mappings of the possible integer values of "sum_type" on Advisory packages to their user-facing
# string representation. Should mirror the createrepo_c source code:
# https://github.com/rpm-software-management/createrepo_c/blob/master/src/checksum.h#L43-L54
# Please keep dict consistent with createrepo_c mapping mentioned above.
ADVISORY_SUM_TYPE_TO_NAME = {
    0: "unknown",
    1: "md5",
    2: "sha",
    3: "sha1",
    4: "sha224",
    5: "sha256",
    6: "sha384",
    7: "sha512"
}



import json
import os

location = os.path.dirname(os.path.realpath(__file__))

with open(os.path.join(location, "copy_config.json")) as copy_config_json:
    COPY_CONFIG_SCHEMA = json.load(copy_config_json)



from .advisory import UpdateCollection, UpdateCollectionPackage, UpdateRecord, UpdateReference  # noqa
from .comps import PackageCategory, PackageEnvironment, PackageGroup, PackageLangpacks  # noqa
from .custom_metadata import RepoMetadataFile  # noqa
from .distribution import Addon, Checksum, DistributionTree, Image, Variant  # noqa
from .modulemd import Modulemd, ModulemdDefaults  # noqa
from .package import Package  # noqa
from .repository import RpmDistribution, RpmPublication, RpmRemote, RpmRepository  # noqa



import unittest
import requests

from pulp_smash import api, config
from pulp_smash.pulp3.utils import (
    gen_distribution,
    gen_repo,
)

from pulp_rpm.tests.functional.utils import (
    gen_rpm_client,
    monitor_task,
)

from pulpcore.client.pulp_rpm import (
    DistributionsRpmApi,
    RepositoriesRpmApi,
    PublicationsRpmApi,
    RpmRpmPublication,
)


class ContentHandlerTests(unittest.TestCase):
    """Whether the RpmDistribution.content_handler* methods work."""

    @classmethod
    def tearDownClass(cls) -> None:
        """Clean up after testing."""
        for f, x in cls.cleanUp:
            f(x)

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the class."""
        cls.cfg = config.get_config()
        cls.api_client = api.Client(cls.cfg, api.json_handler)
        cls.client = gen_rpm_client()
        cls.repo_api = RepositoriesRpmApi(cls.client)
        cls.publications_api = PublicationsRpmApi(cls.client)
        cls.distributions_api = DistributionsRpmApi(cls.client)

        cls.cleanUp = list()

        repo = cls.repo_api.create(gen_repo())
        cls.cleanUp.append((cls.repo_api.delete, repo.pulp_href))

        publish_data = RpmRpmPublication(repository=repo.pulp_href)
        publish_response = cls.publications_api.create(publish_data)
        created_resources = monitor_task(publish_response.task)
        publication_href = created_resources[0]
        cls.cleanUp.append((cls.publications_api.delete, publication_href))

        dist_data = gen_distribution(publication=publication_href)
        dist_response = cls.distributions_api.create(dist_data)
        created_resources = monitor_task(dist_response.task)
        cls.dist = cls.distributions_api.read(created_resources[0])
        cls.cleanUp.append((cls.distributions_api.delete, cls.dist.pulp_href))

    def testConfigRepoInListingUnsigned(self):
        """Whether the served resources are in the directory listing."""
        resp = requests.get(self.dist.base_url)

        self.assertEqual(resp.status_code, 200)
        self.assertIn(b'config.repo', resp.content)
        self.assertNotIn(b'public.key', resp.content)

    def testConfigRepoUnsigned(self):
        """Whether config.repo can be downloaded and has the right content."""
        resp = requests.get(f'{self.dist.base_url}config.repo')

        self.assertEqual(resp.status_code, 200)
        self.assertIn(bytes(f'[{self.dist.name}]\n', 'utf-8'), resp.content)
        self.assertIn(bytes(f'baseurl={self.dist.base_url}\n', 'utf-8'), resp.content)
        self.assertIn(bytes('gpgcheck=0\n', 'utf-8'), resp.content)
        self.assertIn(bytes('repo_gpgcheck=0', 'utf-8'), resp.content)



import urllib.parse
from logging import getLogger

from aiohttp.web_response import Response
from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.db import (
    models,
    transaction,
)

from pulpcore.plugin.models import (
    AsciiArmoredDetachedSigningService,
    CreatedResource,
    Remote,
    Repository,
    RepositoryVersion,
    Publication,
    PublicationDistribution,
    Task,
)
from pulpcore.plugin.repo_version_utils import remove_duplicates, validate_repo_version

from pulp_rpm.app.constants import CHECKSUM_CHOICES
from pulp_rpm.app.models import (
    DistributionTree,
    Package,
    PackageCategory,
    PackageGroup,
    PackageEnvironment,
    PackageLangpacks,
    RepoMetadataFile,
    Modulemd,
    ModulemdDefaults,
    UpdateRecord,
)

log = getLogger(__name__)


class RpmRepository(Repository):
    """
    Repository for "rpm" content.

    Fields:

        sub_repo (Boolean):
            Whether is sub_repo or not
        last_sync_revision_number (Text):
            The revision number
        last_sync_remote (Remote):
            The remote used for the last sync
        last_sync_repo_version (Integer):
            The repo version number of the last sync
        original_checksum_types (JSON):
            Checksum for each metadata type
    """

    TYPE = "rpm"
    CONTENT_TYPES = [
        Package, UpdateRecord,
        PackageCategory, PackageGroup, PackageEnvironment, PackageLangpacks,
        RepoMetadataFile, DistributionTree,
        Modulemd, ModulemdDefaults
    ]

    metadata_signing_service = models.ForeignKey(
        AsciiArmoredDetachedSigningService,
        on_delete=models.SET_NULL,
        null=True
    )
    sub_repo = models.BooleanField(default=False)
    last_sync_revision_number = models.CharField(max_length=20, null=True)
    last_sync_remote = models.ForeignKey(Remote, null=True, on_delete=models.SET_NULL)
    last_sync_repo_version = models.PositiveIntegerField(default=0)
    original_checksum_types = JSONField(default=dict)

    def new_version(self, base_version=None):
        """
        Create a new RepositoryVersion for this Repository.

        Creation of a RepositoryVersion should be done in a RQ Job.

        Args:
            repository (pulpcore.app.models.Repository): to create a new version of
            base_version (pulpcore.app.models.RepositoryVersion): an optional repository version
                whose content will be used as the set of content for the new version

        Returns:
            pulpcore.app.models.RepositoryVersion: The Created RepositoryVersion

        """
        with transaction.atomic():
            version = RepositoryVersion(
                repository=self,
                number=int(self.next_version),
                base_version=base_version)
            version.save()

            if base_version:
                # first remove the content that isn't in the base version
                version.remove_content(version.content.exclude(pk__in=base_version.content))
                # now add any content that's in the base_version but not in version
                version.add_content(base_version.content.exclude(pk__in=version.content))

            if Task.current() and not self.sub_repo:
                resource = CreatedResource(content_object=version)
                resource.save()
            return version

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"

    def finalize_new_version(self, new_version):
        """
        Ensure there are no duplicates in a repo version and content is not broken.

        Remove duplicates based on repo_key_fields.
        Ensure that modulemd is added with all its RPMs.
        Ensure that modulemd is removed with all its RPMs.
        Resolve advisory conflicts when there is more than one advisory with the same id.

        Args:
            new_version (pulpcore.app.models.RepositoryVersion): The incomplete RepositoryVersion to
                finalize.
        """
        if new_version.base_version:
            previous_version = new_version.base_version
        else:
            try:
                previous_version = new_version.previous()
            except RepositoryVersion.DoesNotExist:
                previous_version = None

        remove_duplicates(new_version)

        from pulp_rpm.app.modulemd import resolve_module_packages  # avoid circular import
        resolve_module_packages(new_version, previous_version)

        from pulp_rpm.app.advisory import resolve_advisories  # avoid circular import
        resolve_advisories(new_version, previous_version)
        validate_repo_version(new_version)


class RpmRemote(Remote):
    """
    Remote for "rpm" content.
    """

    TYPE = 'rpm'

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"


class RpmPublication(Publication):
    """
    Publication for "rpm" content.
    """

    TYPE = 'rpm'
    metadata_checksum_type = models.CharField(choices=CHECKSUM_CHOICES, max_length=10)
    package_checksum_type = models.CharField(choices=CHECKSUM_CHOICES, max_length=10)

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"


class RpmDistribution(PublicationDistribution):
    """
    Distribution for "rpm" content.
    """

    TYPE = 'rpm'
    repository_config_file_name = 'config.repo'

    def content_handler(self, path):
        """Serve config.repo and public.key."""
        if path == self.repository_config_file_name:
            val = f"""[{self.name}]
enabled=1
baseurl={settings.CONTENT_ORIGIN}{settings.CONTENT_PATH_PREFIX}{self.base_path}/
gpgcheck=0
"""
            repository_pk = self.publication.repository.pk
            repository = RpmRepository.objects.get(pk=repository_pk)
            signing_service = repository.metadata_signing_service
            if signing_service is None:
                val += 'repo_gpgcheck=0'
            else:
                gpgkey_path = urllib.parse.urljoin(
                    settings.CONTENT_ORIGIN, settings.CONTENT_PATH_PREFIX
                )
                gpgkey_path = urllib.parse.urljoin(gpgkey_path, self.base_path, True)
                gpgkey_path += '/repodata/public.key'

                val += f"""repo_gpgcheck=1
gpgkey={gpgkey_path}
"""
            return Response(body=val)

    def content_handler_list_directory(self, rel_path):
        """Return the extra dir entries."""
        retval = set()
        if rel_path == '':
            retval.add(self.repository_config_file_name)
        return retval

    class Meta:
        default_related_name = "%(app_label)s_%(model_name)s"
